"""A simple template engine.

Shrubbery intends to be the world's easiest template language. It
has two major advantages::

1. The user doesn't need to learn a new syntax. Templates are simply
   structural, and contain no code at all. Logic is dictated by the
   data following an implicit -- though intuitive! -- algorithm.

2. Templates don't have to be valid HTML/XHTML. Shrubbery uses the
   wonderful BeautifulSoup module for the heavy lifting, and can
   handle templates with broken HTML.
"""

from template import Template



from shrubbery.template import Template

template = Template("""<html>
<head> <title>{title}</title> </head>

<body> <h1>{title}</h1>

    <div id="{entry.id}"> <h2>{entry.title}</h2>
        <div class="{entry.content.type}">{entry.content.content}</div>
        <p class="footnote">Updated on {entry.updated}</p>
        <ul>
            <li class="{entry.category}"><a href="http://example.com/{entry.category.term}">{entry.category.label}</a></li>
        </ul>
    </div>
</body>
</html>""")

data = {"entry": [{'category': [{'term': 'weblog', 'label': 'Weblog'},
                                {'term': 'python', 'label': 'Python stuff'}],
                   'content': {'content': '<p>This is my first post!</p>', 'type': 'html'},
                   'id': '1', 
                   'title': 'First post',
                   'updated': '2007-01-23T18:18:43Z'},
                  {'content': {'content': '<p>This is the second post...</p>', 'type': 'html'},
                   'id': '0', 
                   'title': 'One more post',
                   'updated': '2007-01-18T13:31:43Z'}],
        "title": "This is the title"}

print template.process(data, escape=False).prettify()



"""
Shrubbery is a Smart Html Renderer (Using Blocks to Bind Expressions
RepeatedlY).

This is a simple templating engine designed to convert JSON_ data
into HTML or XML. The idea behind it is that templates are very
simple, containing only HTML tags, with nodes being repeated as
required by the number of elements in the data.

_JSON: http://json.org/

"""
import re
import cgi
import sgmllib

# Monkeypatch sgmllib to avoid breaking ampersands.
sgmllib.entityref = re.compile("[^\W\w]")

from BeautifulSoup import BeautifulSoup, NavigableString


# Basic regexp for matching all expressions in a template.
TOKEN = r"([\w_:-]+)(\[\d+\])*"
EXPR = re.compile(r"\{(%s(\.%s)*)\}" % (TOKEN, TOKEN))


class Template(BeautifulSoup):
    """A Shrubbery template.

    Shrubbery is a smart template that holds *no logic* whatsoever,
    with behavior being dictated by the data.

    """

    def process(self, data, remove_empty_nodes=True, escape=True):
        """Process the template given some data.

        Here, ``data`` should be a dict-like object with the variables
        to be replaced. If ``escape`` is true, any HTML in the data
        will be escaped before the replacement.

        The option ``remove_empty_nodes`` will remove nodes with
        replacements not present in ``data``.
        """
        # Get all namespaces available in the template.
        nstree = _get_namespaces(self)

        # Replace tree with elements from ``data``.
        tree = search(self.copy(), data, '', nstree,
                remove_empty_nodes, escape)

        # Fix tags in replacements.
        tree = apply(tree, "%SHRUBBERY_OPEN_TAG%", "{")
        tree = apply(tree, "%SHRUBBERY_CLOSE_TAG%", "}")

        return tree
    render = process

    def copy(self):
        """A lazy and *costy* copy of the template."""
        return Template(unicode(self))


def search(tree, data, ns, nstree, remove_empty_nodes, escape):
    """
    Search for all expressions inside a given namespace.

    This function finds all expressions inside a given namespace
    (say, ``entry`` or ``collection.id``). After that it locates
    the uppermost node common to all nodes, and replicates it
    according to the number of elements in the applied data. Each
    node is then processed to replace the expressions with the
    proper values.

    """
    # Search for all expressions inside a given namespace.
    regexp = re.compile(r"\{%s.*\}" % re.escape(ns.rstrip('.')))
    nodes = tree.findAll(replaceable(regexp)) + tree.findAll(text=regexp)
    if not nodes: return tree

    # Get the node in common, its parent and position.
    node = find_common_node(nodes)
    parent = node.parent
    index = parent.contents.index(node)

    # Now for every instance of ``data`` we will clone the
    # parent node, replacing the appropriate data values.
    # First a little cleanup...
    if not isinstance(data, list): data = [data]
    data = [v for v in data if v not in [None, {}, []]]
    for i, values in enumerate(data):
        # Return a node with the values applied.
        repl = apply(clone(node),
                "\{%s\}" % re.escape(ns.rstrip('.')),
                values, escape)
        # Continue replacement in deeper namespaces.
        keys = nstree.keys()
        keys.sort()
        keys.reverse()  # make indexed ns be replaced first.
        for k in keys:
            new_ns = '%s%s.' % (ns, k)
            new_nstree = nstree.get(k)

            # Get any indexes.
            indexes = re.match(TOKEN, k)
            k, slice_ = indexes.groups()

            # Get subdata for this node.
            try:
                subdata = values.get(k)
            except:
                subdata = getattr(values, k, None)
            if slice_ and subdata:
                subdata = subdata[int(slice_[1:-1])]

            repl = search(repl, subdata, new_ns,
                    new_nstree, remove_empty_nodes, escape)
        parent.insert(index+i, repl)

    # Remove this node from parent.
    if data or remove_empty_nodes: node.extract()
    return tree


def apply(tree, match, data, escape=False):
    """
    Apply replacement in expressions.

    This function replaces the expressions with the values in ``data``.

    """
    # Find all replaceable nodes from this namespace.
    regexp = re.compile(match)
    nodes = tree.findAll(replaceable(regexp)) + tree.findAll(text=regexp)

    # Normalize data. Data can be a dict or object when it's used
    # as a stub for defining the topmost element to repeat.
    if isinstance(data, (int, long, float)):
        data = unicode(data)
    elif isinstance(data, str):
        data = data.decode('utf-8')
    elif not isinstance(data, unicode):
        data = u''

    # Escape shrubbery tags inside replacement.
    data = EXPR.sub(lambda m:
            "%%SHRUBBERY_OPEN_TAG%%%s%%SHRUBBERY_CLOSE_TAG%%"
            % m.group(1), data)

    # Do the replacement.
    if escape: text = cgi.escape(data)
    else: text = data
    for node in nodes:
        if isinstance(node, NavigableString):
            repl = regexp.sub(text, node.string)
            if node.parent: node.replaceWith(repl)
        else:
            if node.string:
                node.string = regexp.sub(text, node.string)
            for k, v in node.attrs:
                node[k] = regexp.sub(data, v).strip()
                if not node[k]: del node[k]
    return tree


def find_common_node(nodes):
    """
    Return the lowest node in common given a list of nodes.

    Given the nodes in ``nodes``, this function returns the
    lowest parent in common to all nodes.

    """
    # Senseless for a single node.
    if len(nodes) == 1: return nodes[0]

    # Get a list of all parents.
    parents = [[node] + node.findParents() for node in nodes]
    # Here we use the smallest list, iterating over each element
    # and checking if the candidate appears in all list of
    # parents.
    parents.sort(key=len)
    for candidate in parents.pop(0):
        for other_parents in parents:
            if candidate not in other_parents: break
        else:
            return candidate


def _get_namespaces(soup):
    """
    Get all namespaces from the template.

    This function searches ``soup`` for all namespaces, and returns
    them as a nested dict. A template like this::

        <html>
            <div id="{entry.id}"></div>
        </html>

    Would return ``{u'entry': {u'id': {}}}``, eg.

    """
    namespaces = {}
    nodes = soup.findAll(replaceable(EXPR)) + soup.findAll(text=EXPR)
    for node in nodes:
        for ns in EXPR.findall(node.string or ''):
            ns = ns[0].split('.')
            root = namespaces
            for item in ns:
                root = root.setdefault(item, {})
        if hasattr(node, 'attrs'):
            for k, v in node.attrs:
                for ns in EXPR.findall(v):
                    ns = ns[0].split('.')
                    root = namespaces
                    for item in ns:
                        root = root.setdefault(item, {})
    return namespaces


def clone(node):
    """
    Clone a node.

    We simply parse the HTML from ``node`` to be copied. Yes, I'm lazy.

    """
    return Template(unicode(node))


def replaceable(regexp=EXPR):
    """
    Return function to find nodes that are replaceable in a given namespace.

    Given a namespace defined by ``regexp``, this function returns
    a callable that can be used in ``findAll`` to search for
    corresponding nodes.

    """

    def func(tag):
        """Check both text and attributes for ``regexp``."""
        if tag.string and regexp.search(tag.string): return True
        for (k, v) in tag.attrs:
            if regexp.search(v): return True
        return False
    return func



from shrubbery.template import Template

template = Template("""<?xml version="1.0" encoding="utf-8"?>
<feed xmlns="http://www.w3.org/2005/Atom">

    <title>{title}</title>
    <link rel="self" href="{ROOT}atom"/>
    <updated>{entries[0].updated}</updated>
    <author>
        <name>{author.name}</name>
        <email>{author.email}</email>
        <uri>{author.uri}</uri>
    </author>
    <id>{ROOT}</id>

    <entry>
        <id>{entries.href}</id>
        <title>{entries.title}</title>
        <updated>{entries.updated}</updated>

        <content type="xhtml">
            <div xmlns="http://www.w3.org/1999/xhtml">
                {entries.content}
            </div>
        </content>
        <link rel="alternate" href="{entries.href}"/>
        <summary>{entries.summary}</summary>

        <category term="{entries.category}"></category>
        <published>{entries.published}</published>
    </entry>
</feed>""")

data = {'title'  : 'A simple Atom feed',
        'ROOT'   : 'http://example.com/',
        'author' : {'name': 'John Doe', 'email': 'jdoe@example.com', 'uri': 'http://djoe.example.com/'},
        'entries': [{
            'href'     : 'http://example.com/1',
            'title'    : 'Second post!',
            'updated'  : '2007-08-10T13:29:14Z',
            'content'  : '<p>This is my second post!</p>',
            'summary'  : 'My second post',
            'category' : 'blog',
            'published': '2007-08-10T13:29:14Z'}, {

            'href'     : 'http://example.com/0',
            'title'    : 'First post!',
            'updated'  : '2007-08-10T13:28:04Z',
            'content'  : '<p>This is my first post!</p>',
            'summary'  : 'My first post',
            'category' : ['blog', 'test'],
            'published': '2007-08-10T13:28:04Z'}]}
                
print template.process(data, escape=False)



from shrubbery.template import Template

def test_basic():
    # Simple basic tests of substitution.
    assert str(Template("{content}").process({})) == ''
    assert str(Template("{content}").process({"content": None})) == ''
    assert str(Template("{content}").process({"content": ""})) == ''
    assert str(Template("{content}").process({"content": "Hello, World!"})) == 'Hello, World!'
    assert str(Template("{content}").process({"content": ["Hello", ",", " ", "World!"]})) == 'Hello, World!'
    assert str(Template("I {verb} {noun}.").process({"verb": "like", "noun": "Python"})) == 'I like Python.'
    assert str(Template("{var1}").process({"var1": "Hello, World!"})) == 'Hello, World!'
    assert str(Template("{var_1}").process({"var_1": "Hello, World!"})) == 'Hello, World!'


def test_nested_dict():
    # Tests using nested data and the dot idiom.
    assert str(Template("I {entry.verb} {entry.noun}.").process({"entry": {"verb": "like", "noun": "Python"}})) == \
    'I like Python.'


def test_html():
    # Testing how HTML nodes are handled.
    assert str(Template("""
        <a href="{url}">{name}</a>
    """).process(
        {"url": "http://google.com/", "name": "Google"}
    )).strip() == '<a href="http://google.com/">Google</a>'
    
    assert str(Template("""
        <a href="{url}">{name}</a>
    """).process(
        [{"url": "http://google.com/", "name": "Google"},
         {"url": "http://yahoo.com/", "name": "Yahoo!"},
    ])).strip() == '<a href="http://google.com/">Google</a><a href="http://yahoo.com/">Yahoo!</a>'

    assert str(Template("""
        <li><a href="{url}">{name}</a></li>
    """).process(
        [{"url": "http://google.com/", "name": "Google"},
         {"url": "http://yahoo.com/", "name": "Yahoo!"},
    ])).strip() == '<li><a href="http://google.com/">Google</a><a href="http://yahoo.com/">Yahoo!</a></li>'

    assert str(Template("""
        <li><a href="{entry.url}">{entry.name}</a></li>
    """).process(
        {"entry": [{"url": "http://google.com/", "name": "Google"},
                   {"url": "http://yahoo.com/", "name": "Yahoo!"},
    ]})).strip() == '<li><a href="http://google.com/">Google</a><a href="http://yahoo.com/">Yahoo!</a></li>'

    # If we want to repeat the ``<li>`` in the last example we can create a dummy attribute.
    assert str(Template("""
        <li class="{entry}"><a href="{entry.url}">{entry.name}</a></li>
    """).process(
        {"entry": [{"url": "http://google.com/", "name": "Google"},
                   {"url": "http://yahoo.com/", "name": "Yahoo!"},
    ]})).strip() == '<li><a href="http://google.com/">Google</a></li><li><a href="http://yahoo.com/">Yahoo!</a></li>'


def test_meta_replacements():
    # Nodes in the replacements should be preserved.
    assert str(Template("{a} and {b}").process({"a": "{b}", "b": "foo"})) == '{b} and foo'
    assert str(Template("{a} and {b}").process({"a": "foo", "b": "{a}"})) == 'foo and {a}'


def test_ampersand_escaping():
    # Ampersands should not be escaped.
    assert str(Template("<p>{http://example.com/?a=1&b=2}</p>")) == '<p>{http://example.com/?a=1&b=2}</p>'
    assert str(Template("<p>2006&mdash;2007</p>")) == '<p>2006&mdash;2007</p>'


def test_misc():
    assert str(Template("""
        <html>I love {a} a lot, it's {b}!</html>
    """).process(
        {"a": "sushi", "b": "nice"}
    )).strip() == "<html>I love sushi a lot, it's nice!</html>"

    assert str(Template("""
        <html>I love {a} a lot, it's <span>{b}</span>!</html>
    """).process(
        {"a": "sushi", "b": "nice"}
    )).strip() == "<html>I love sushi a lot, it's <span>nice</span>!</html>"

    assert str(Template("""
        <html>I love {a} a lot, it's {b}!</html>
    """).process(
        {"a": "{b}", "b": "nice"}
    )).strip() == "<html>I love {b} a lot, it's nice!</html>"

def test_escape():
    assert str(Template("""
        <p>{content}</p>
    """).process(
        {"content": "This is content <em>with</em> HTML."}
    )).strip() == "<p>This is content &lt;em&gt;with&lt;/em&gt; HTML.</p>"

    assert str(Template("""
        <p>{content}</p>
    """).process(
        {"content": "This is content <em>with</em> HTML."},
        escape=False
    )).strip() == "<p>This is content <em>with</em> HTML.</p>"

def test_objects():
    class container(object): pass
    entry = container()
    entry.verb = "like"
    entry.noun = "Python"
    assert str(Template("I {entry.verb} {entry.noun}.").process({"entry": entry})) == \
    'I like Python.'

def test_index():
    assert str(Template("""
        <georss:point>{geometry.coordinates[0]} {geometry.coordinates[1]}</georss:point>
    """).process(
        {"geometry": { "type": "Point", "coordinates": [45.256, -71.92] }}
    )).strip() == "<georss:point>45.256 -71.92</georss:point>"

    template = Template("""<feed><updated>{entries[0].updated}</updated><entry><id>{entries.id}</id><updated>{entries.updated}</updated></entry></feed>""")
    assert str(template.process({"entries": [{"updated": "today", "id": 1}, {"updated": "yesterday", "id": 0}]})) == \
    '<feed><updated>today</updated><entry><id>1</id><updated>today</updated></entry><entry><id>0</id><updated>yesterday</updated></entry></feed>'

    template = Template("""<feed><updated>{members[0].entity.updated}</updated><entry><id>{members.id}</id><updated>{members.entity.updated}</updated></entry></feed>""")
    assert str(template.process({"members": [{'entity': {"updated": "today"}, "id": 1}, {'entity': {"updated": "yesterday"}, "id": 0}]})) == \
    '<feed><updated>today</updated><entry><id>1</id><updated>today</updated></entry><entry><id>0</id><updated>yesterday</updated></entry></feed>'
    template = Template("""<feed><updated>{entries[0].entity.updated}</updated><entry><id>{entries.id}</id><updated>{entries.entity.updated}</updated></entry></feed>""")
    assert str(template.process({"entries": [{'entity': {"updated": "today"}, "id": 1}, {'entity': {"updated": "yesterday"}, "id": 0}]})) == \
    '<feed><updated>today</updated><entry><id>1</id><updated>today</updated></entry><entry><id>0</id><updated>yesterday</updated></entry></feed>'



RELEASE = True

from setuptools import setup, find_packages
import sys, os

classifiers = """\
Development Status :: 4 - Beta
Intended Audience :: Developers
License :: OSI Approved :: MIT License
Operating System :: OS Independent
Programming Language :: Python
Topic :: Text Processing :: Markup
"""

version = '0.3.3'

setup(name='shrubbery',
      version=version,
      description="Simple and smart template engine to generate HTML/XML",
      long_description='''\
Shrubbery is a *Smart Html Renderer Using Blocks to Bind Expressions Repeatedly*. You can also think of it as the "world's easiest templating engine": Templates hold no logic whatsoever, with nodes being repeated implicitely as needed by the replacement data.
''',
      classifiers=filter(None, classifiers.split("\n")),
      keywords='json template html xml',
      author='Roberto De Almeida',
      author_email='roberto@dealmeida.net',
      url='http://taoetc.org/46',
      download_url = "http://cheeseshop.python.org/packages/source/s/shrubbery/shrubbery-%s.tar.gz" % version,
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      test_suite = 'nose.collector',
      include_package_data=True,
      zip_safe=True,
      install_requires=[
          # -*- Extra requirements: -*-
          "BeautifulSoup",
      ],
      entry_points="""
      # -*- Entry points: -*-
      [python.templating.engines]
      shrubbery = shrubbery.buffet:ShrubberyTemplatePlugin
      """,
      )
     



from shrubbery.template import Template

template = Template("""<div class="vcard">
    <img style="float:left; margin-right:4px" src="{photourl}" alt="photo" class="photo"/>

    <a class="url fn n" href="{url}">  <span class="givenname">{givenname}</span>
        <span class="additionalname">{additionalname}</span>

        <span class="familyname">{familyname}</span>
    </a>

    <div class="org">{org}</div>

    <a class="email" href="mailto:{email}">{email}</a>

    <div class="adr">
        <div class="streetaddress">{streetaddress}</div>

        <span class="locality">{locality}</span>, 
        <span class="region">{region}</span>, 
        <span class="postalcode">{postalcode}</span>

        <span class="countryname">{countryname}</span>
    </div>

    <div class="tel">{tel}</div>

    <ul><li class="{im}"><a class="url" href="{im.href}">{im.protocol}</a></li></ul>
</div>""")

data = {'additionalname': 'Ferreira',
        'countryname': 'Brazil',
        'email': 'roberto@dealmeida.net',
        'familyname': 'De Almeida',
        'givenname': 'Roberto Antonio',
        'im': [{'href': 'jabber:roberto@dealmeida.net', 'protocol': 'Jabber'},
               {'href': 'skype:robertodealmeida', 'protocol': 'Skype'}],
        'locality': 'Cachoeira Paulista',
        'org': 'CPTEC/INPE',
        'photourl': 'http://dealmeida.net/roberto-t.jpg',
        'postalcode': '12630-000',
        'region': 'SP',
        'streetaddress': 'Rod. Pres. Dutra km 40',
        'tel': '+55 12 3456 7890',
        'url': 'http://dealmeida.net'}

print template.process(data).prettify()



"""
Plugin for Buffet, modified from BuffetString.

ShrubberyTemplatePlugin was tested with Pylons_, and it should work
fine. Just remember that shrubbery *cannot* call functions, only
attributes.

_Pylons: http://pylonshq.com/
"""

import os

from shrubbery.template import Template


class ShrubberyTemplatePlugin(object):
    """This is a simple Buffet template for Shrubbery."""

    # all template plugins need to define a default file extension
    extension = "shrb"

    def __init__(self, extra_vars_func=None, options=None):
        """extra_vars_func == optional callable() that returns a dict
        options == optional dict() of configuration settings
        """
        self.get_extra_vars = extra_vars_func
        if options:
            self.options = options
        else:
            self.options = dict()

    def load_template(self, template_name):
        """template_name == dotted.path.to.template (without .ext)

        The dotted notation is present because many template engines
        allow templates to be compiled down to Python modules.  TurboGears
        uses that feature to its adavantage, and for ease of integration
        the python.templating.engines plugin format requires the path to
        the template to be supplied as a dotted.path.to.template regardless
        of whether is is a module or not.

        In the case of shrubbery templates, they are just simple
        text files, so we deal with the dotted notation and translate
        it into a standard file path to open the text file.
        """

        # split the template name up on the dots
        parts = template_name.split('.')

        # get the filename.ext
        template_filename = "%s.%s" % (parts.pop(), self.extension)

        # get the _real_ path to the template text file
        template_path = os.path.join(*parts)
        template_file_path = os.path.join(template_path, template_filename)

        template_file = open(template_file_path)
        template_obj = Template(template_file.read())
        template_file.close()

        # Apply configuration to template.
        template_obj = template_obj.process(self.options,
                remove_empty_nodes=False)

        return template_obj

    def render(self, info, format="html", fragment=False, template=None):
        """info == dict() of variables to stick into the template namespace
        format == output format if applicable
        fragment == special rules about rendering part of a page
        template == dotted.path.to.template (without .ext)

        You might not need all of these arguments.  info and template are the
        only ones used in this simple example.
        """

        # check to see if we were passed a function get extra vars
        if callable(self.get_extra_vars):
            info.update(self.get_extra_vars())

        template_obj = self.load_template(template)
        return template_obj.process(info, remove_empty_nodes=True)
